package visitor;

import datapackage.*;
import mainpackage.*;
import syntaxtree.*;
import java.util.ArrayList;
import java.util.List;

public class MethodVisitor extends GJDepthFirst<String, Function>
{

    private MJClass currentClass;
    private boolean returnTypeMode = false;
    private List<String> argTypeList;

    @Override
    public String visit(MainClass n, Function argu) throws Exception
    {
        currentClass = Main.mainClass;
          
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
        n.f7.accept(this, argu);
        n.f8.accept(this, argu);
        n.f9.accept(this, argu);
        n.f10.accept(this, argu);
        n.f11.accept(this, argu);
        n.f12.accept(this, argu);
        n.f13.accept(this, argu);

        n.f14.accept(this, Main.mainClass.getFunctions().get("main"));

        returnTypeMode = true;
        n.f15.accept(this, Main.mainClass.getFunctions().get("main"));
        returnTypeMode = false;

        n.f16.accept(this, argu);
        n.f17.accept(this, argu);

        return null;
    }

    @Override
    public String visit(ClassDeclaration n, Function argu) throws Exception
    {
        currentClass = Main.programClasses.get(n.f1.f0.tokenImage);

        return super.visit(n, argu);
    }

    @Override
    public String visit(ClassExtendsDeclaration n, Function argu) throws Exception
    {
        currentClass = Main.programClasses.get(n.f1.f0.tokenImage);

        return super.visit(n, argu);
    }

    @Override
    public String visit(VarDeclaration n, Function argu) throws Exception
    {
        String type = n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);

        if (argu != null)
        {
            //If another variable exists with the same name in the function
            //or an argument of the function has the same name or the type isn't correct
            if (argu.getVariables().containsKey(n.f1.f0.tokenImage)
                || argu.getArguments().containsKey(n.f1.f0.tokenImage)
                || !type.equals("array") && !type.equals("int") && !type.equals("boolean")
                   && !Main.programClasses.containsKey(type) && !Main.mainClass.getIdentifier().equals(type))
            {
                System.err.println("Another variable exists with the same name in the function "
                                   + "or an argument of the function has the same name or the type isn't correct");
                throw new TypeCheckingError();
            }

            argu.getVariables().put(n.f1.f0.tokenImage, new Variable(n.f1.f0.tokenImage, type));
        }

        return null;
    }

    @Override
    public String visit(MethodDeclaration n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
                             
        n.f7.accept(this, currentClass.getFunctions().get(n.f2.f0.tokenImage));        
        
        returnTypeMode = true;
        n.f8.accept(this, currentClass.getFunctions().get(n.f2.f0.tokenImage));
        returnTypeMode = false;
        
        n.f9.accept(this, argu);
        
        returnTypeMode = true;
        String exp = n.f10.accept(this, currentClass.getFunctions().get(n.f2.f0.tokenImage));
        returnTypeMode = false;
        //If the return type of the expression is different from the return type of the function
        if(!currentClass.getFunctions().get(n.f2.f0.tokenImage).getType().equals(exp))
        {
            System.err.println("The returned expression is different from the return type of the function!");
            throw new TypeCheckingError();            
        }
        
        n.f11.accept(this, argu);
        n.f12.accept(this, argu);

        return null;
    }    

    @Override
    public String visit(ArrayType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "array";
    }

    @Override
    public String visit(BooleanType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "boolean";
    }

    @Override
    public String visit(IntegerType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "int";
    }

    @Override
    public String visit(AssignmentStatement n, Function argu) throws Exception
    {        
        returnTypeMode = false;
        n.f0.accept(this, argu);
        returnTypeMode = true;
         
        //If the identifier used for this assignment doesn't exist
        if (!argu.getVariables().containsKey(n.f0.f0.tokenImage)
            && !argu.getArguments().containsKey(n.f0.f0.tokenImage)
            && !currentClass.getVariables().containsKey(n.f0.f0.tokenImage))
        {
            //if the variable found at an extended class that's not the main do nothing
            if(currentClass.getExtendsClass() != null 
               && !Main.mainClass.getIdentifier().equals(currentClass.getExtendsClass())
               && Main.programClasses.get(currentClass.getExtendsClass()).getVariables().containsKey(n.f0.f0.tokenImage))
            {                
            }
            else
            {
                System.err.println("The identifier used for this assignment doesn't exist!");
                throw new TypeCheckingError();                
            }
        }
       
        n.f1.accept(this, argu);                 

        String exp = n.f2.accept(this, argu);        
        //If the type of the result of the expression doesn't match
        if (!getVar(n.f0.f0.tokenImage, argu).getType().equals(exp))
        {
            //if the argument is an extended class of the original do nothing
            if(!Main.mainClass.getIdentifier().equals(exp)
               && Main.programClasses.containsKey(exp)
               && Main.programClasses.get(exp).getExtendsClass() != null
               && Main.programClasses.get(exp).getExtendsClass().equals(getVar(n.f0.f0.tokenImage, argu).getType()))
            {

            }
            else
            {
                System.err.println("Assignment between different types!");
                throw new TypeCheckingError();                        
            }                        
        }
        
        n.f3.accept(this, argu);        
        
        return null;
    }  

    @Override
    public String visit(ArrayAssignmentStatement n, Function argu) throws Exception
    {        
        returnTypeMode = false;
        n.f0.accept(this, argu);
        returnTypeMode = true;
        //If the identifier used for this assignment doesn't exist or it's not an array
        if (!argu.getVariables().containsKey(n.f0.f0.tokenImage)
            && !argu.getArguments().containsKey(n.f0.f0.tokenImage)
            && !currentClass.getVariables().containsKey(n.f0.f0.tokenImage)
            || !getVar(n.f0.f0.tokenImage, argu).getType().equals("array"))
        {
            System.err.println("The identifier used for this assignment doesn't exist or it's not an array");
            throw new TypeCheckingError();
        }
        
        n.f1.accept(this, argu);
        
        String exp = n.f2.accept(this, argu);
        if(!exp.equals("int"))
        {
            System.err.println("The array index isn't an integer!");
            throw new TypeCheckingError();
        }        
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        
        exp = n.f5.accept(this, argu);
        if(!exp.equals("int"))
        {
            System.err.println("The expression assigned to the array isn't an integer!");
            throw new TypeCheckingError();
        }        
        
        n.f6.accept(this, argu);
        
        return null;
    }

    @Override
    public String visit(IfStatement n, Function argu) throws Exception
    {
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        
        String exp = n.f2.accept(this, argu);
        if(!exp.equals("boolean"))
        {
            System.err.println("The condition of the if statement isn't boolean!");
            throw new TypeCheckingError();
        }
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
        
        return null;
    }

    @Override
    public String visit(WhileStatement n, Function argu) throws Exception
    {
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        
        String exp = n.f2.accept(this, argu);
        if(!exp.equals("boolean"))
        {
            System.err.println("The condition of the while statement isn't boolean!");
            throw new TypeCheckingError();
        }
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        
        return null;
    }

    @Override
    public String visit(PrintStatement n, Function argu) throws Exception
    {
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        
        String exp = n.f2.accept(this, argu);
        if(!exp.equals("int") && !exp.equals("boolean"))
        {
            System.err.println("The expression to print isn't an integer nor a boolean!");
            throw new TypeCheckingError();
        }
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        
        return null;
    }    

    @Override
    public String visit(AndExpression n, Function argu) throws Exception
    {
        String leftClause = n.f0.accept(this, argu);        
        if (!leftClause.equals("boolean"))
        {
            System.err.println("The left clause of the && isn't boolean");
            throw new TypeCheckingError();
        }
        
        n.f1.accept(this, argu);
        
        String rightClause = n.f2.accept(this, argu);
        if (!rightClause.equals("boolean"))
        {
            System.err.println("The right clause of the && isn't boolean");
            throw new TypeCheckingError();
        }

        return "boolean";
    }

    @Override
    public String visit(CompareExpression n, Function argu) throws Exception
    {
        String leftExp = n.f0.accept(this, argu);
        if (!leftExp.equals("int"))
        {
            System.err.println("The left expression of the comparison isn't an integer");
            throw new TypeCheckingError();
        }        
        
        n.f1.accept(this, argu);
        
        String rightExp = n.f2.accept(this, argu);
        if (!rightExp.equals("int"))
        {
            System.err.println("The right expression of the comparison isn't an integer");
            throw new TypeCheckingError();
        }

        return "boolean";
    }

    @Override
    public String visit(PlusExpression n, Function argu) throws Exception
    {
        String leftExp = n.f0.accept(this, argu);
        if (!leftExp.equals("int"))
        {
            System.err.println("The left expression of the addition isn't an integer");
            throw new TypeCheckingError();
        }        
        n.f1.accept(this, argu);
        
        String rightExp = n.f2.accept(this, argu);
        if (!rightExp.equals("int"))
        {
            System.err.println("The right expression of the addition isn't an integer");
            throw new TypeCheckingError();
        }

        return "int";        
    }

    @Override
    public String visit(MinusExpression n, Function argu) throws Exception
    {
        String leftExp = n.f0.accept(this, argu);
        if (!leftExp.equals("int"))
        {
            System.err.println("The left expression of the subtraction isn't an integer");
            throw new TypeCheckingError();
        }        
        
        n.f1.accept(this, argu);
        
        String rightExp = n.f2.accept(this, argu);
        if (!rightExp.equals("int"))
        {
            System.err.println("The right expression of the subtraction isn't an integer");
            throw new TypeCheckingError();
        }

        return "int";
    }

    @Override
    public String visit(TimesExpression n, Function argu) throws Exception
    {
        String leftExp = n.f0.accept(this, argu);
        if (!leftExp.equals("int"))
        {
            System.err.println("The left expression of the multiplication isn't an integer");
            throw new TypeCheckingError();
        }        
        
        n.f1.accept(this, argu);
        
        String rightExp = n.f2.accept(this, argu);
        if (!rightExp.equals("int"))
        {
            System.err.println("The right expression of the multiplication isn't an integer");
            throw new TypeCheckingError();
        }

        return "int";
    }

    @Override
    public String visit(ArrayLookup n, Function argu) throws Exception
    {
        
        String type = n.f0.accept(this, argu);
        if (!type.equals("array"))
        {
            System.err.println("The array reference for the lookup isn't an array!");
            throw new TypeCheckingError();
        }        
        
        n.f1.accept(this, argu);
        
        String index = n.f2.accept(this, argu);
        if (!index.equals("int"))
        {
            System.err.println("The array index of the lookup isn't an integer!");
            throw new TypeCheckingError();
        }        
        
        n.f3.accept(this, argu);

        return "int";
    }

    @Override
    public String visit(ArrayLength n, Function argu) throws Exception
    {               
        String type = n.f0.accept(this, argu);
        if (!type.equals("array"))
        {
            System.err.println("The array reference for the length isn't an array!");
            throw new TypeCheckingError();
        }        
        
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);

        return "int";
    }

    @Override
    public String visit(MessageSend n, Function argu) throws Exception
    {
        String type = n.f0.accept(this, argu);
        if(!Main.programClasses.containsKey(type))
        {
            System.err.println("No such type for message send!");
            throw new TypeCheckingError();
        }
        
        n.f1.accept(this, argu);
        
        returnTypeMode = false;
        n.f2.accept(this, argu);
        returnTypeMode = true;
        //If the function called doesn't exist in the class "type"
        if(!Main.programClasses.get(type).getFunctions().containsKey(n.f2.f0.tokenImage))
        {            
            if(Main.programClasses.get(type).getExtendsClass() != null
               && !Main.mainClass.getIdentifier().equals(Main.programClasses.get(type).getExtendsClass()))
            {
                //And if it isn't a function of an extended class
                if(!Main.programClasses.get(Main.programClasses.get(type).getExtendsClass())
                   .getFunctions().containsKey(n.f2.f0.tokenImage))
                {
                    System.err.println("The function called doesn't exist in the requested class nor the extended!");
                    throw new TypeCheckingError();                    
                }
                else
                {
                    type = Main.programClasses.get(Main.programClasses.get(type).getExtendsClass()).getIdentifier();
                }
            }
            else
            {
                System.err.println("The function called doesn't exist in the requested class!");
                throw new TypeCheckingError();                
            }                        
        }        
        
        n.f3.accept(this, argu);
        
        argTypeList = new ArrayList();
        n.f4.accept(this, argu);    
        //check if the called function doesn't have the same number of arguments with original
        if(Main.programClasses.get(type).getFunctions().get(n.f2.f0.tokenImage).getArguments().size()
           != argTypeList.size())
        {      
            System.err.println("The called function doesn't have the same number of arguments with original!");
            throw new TypeCheckingError();
        }
        else
        {                                                               
            List<Variable> origArgList = new ArrayList();
            origArgList.addAll(Main.programClasses.get(type).getFunctions()
                               .get(n.f2.f0.tokenImage).getArguments().values());

            for(int i = 0; i < origArgList.size(); i++)
            {
                //check if they don't have the same type of arguments
                if(!origArgList.get(i).getType().equals(argTypeList.get(i)))
                {                         
                    //if the argument is an extended class of the original do nothing
                    if(Main.programClasses.get(argTypeList.get(i)).getExtendsClass() != null
                       && Main.programClasses.get(argTypeList.get(i)).getExtendsClass().equals(origArgList.get(i).getType()))
                    {
                        
                    }
                    else
                    {                   
                        System.err.println("The arguments of the called function aren't correct!");
                        throw new TypeCheckingError();                        
                    }                   
                }                            
            }
        }        
        
        n.f5.accept(this, argu);
        
        return Main.programClasses.get(type).getFunctions().get(n.f2.f0.tokenImage).getType();
    }

    @Override
    public String visit(ExpressionList n, Function argu) throws Exception
    {
        String type = n.f0.accept(this, argu); 
        argTypeList.add(type);
        
        n.f1.accept(this, argu);
        
        return null;
    }   
    
    @Override
    public String visit(ExpressionTerm n, Function argu) throws Exception
    {        
        n.f0.accept(this, argu);
        
        String type = n.f1.accept(this, argu);     
        argTypeList.add(type);
        
        return type;
    }    
    
    @Override
    public String visit(IntegerLiteral n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "int";
    }

    @Override
    public String visit(TrueLiteral n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "boolean";
    }

    @Override
    public String visit(FalseLiteral n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "boolean";
    }

    @Override
    public String visit(Identifier n, Function argu) throws Exception
    {
        super.visit(n, argu);

        if(returnTypeMode)
        {            
            String type = getVar(n.f0.tokenImage, argu).getType();            
            return type;
        }

        return n.f0.tokenImage;
    }

    @Override
    public String visit(ThisExpression n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return currentClass.getIdentifier();
    }

    @Override
    public String visit(ArrayAllocationExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        
        String type = n.f3.accept(this, argu);
        if(!type.equals("int"))
        {
            System.err.println("The allocated array has size that's not an integer!");
            throw new TypeCheckingError();
        }
        
        n.f4.accept(this, argu);
        
        return "array";
    }

    @Override
    public String visit(AllocationExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        returnTypeMode = false; 
        n.f1.accept(this, argu);
        returnTypeMode = true;
        
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        
        return n.f1.f0.tokenImage;
    }

    @Override
    public String visit(NotExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        String type = n.f1.accept(this, argu);
        if(!type.equals("boolean"))
        {
            System.err.println("Non boolean expression used for Not!");
            throw new TypeCheckingError();
        }
        
        return "boolean";
    }

    @Override
    public String visit(BracketExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        String type = n.f1.accept(this, argu);
        
        n.f2.accept(this, argu);
        
        return type;
    }   

    public Variable getVar(String identifier, Function argu) throws Exception
    {        
        Variable var = null;

        if(Main.mainClass.getFunctions().get("main").getArguments().containsKey(identifier)
           && currentClass.getIdentifier().equals(Main.mainClass.getIdentifier()))
        {
            System.err.println("The argument of the main function cannot be used!");
            throw new TypeCheckingError();            
        }
        else if (argu.getVariables().containsKey(identifier))
        {
            var = argu.getVariables().get(identifier);
        }
        else if (argu.getArguments().containsKey(identifier))
        {
            var = argu.getArguments().get(identifier);
        }
        else if (currentClass.getVariables().containsKey(identifier))
        {
            var = currentClass.getVariables().get(identifier);
        }
        else if (currentClass.getExtendsClass() != null 
                 && !currentClass.getExtendsClass().equals(Main.mainClass.getIdentifier())
                 && Main.programClasses.get(currentClass.getExtendsClass()).getVariables().containsKey(identifier))
        {
            var = Main.programClasses.get(currentClass.getExtendsClass()).getVariables().get(identifier);
        }                
        else
        {
            System.err.println("The identifier wasn't found!");
            throw new TypeCheckingError();
        }
        
        return var;
    }    
}
